@@ -18,11 +18,6 @@
 #ifndef _LOG4CXX_PATTERN_LAYOUT_H
 #define _LOG4CXX_PATTERN_LAYOUT_H
 
-#if defined(_MSC_VER)
-	#pragma warning ( push )
-	#pragma warning ( disable: 4231 4251 4275 4786 )
-#endif
-
 #include <log4cxx/layout.h>
 #include <log4cxx/pattern/loggingeventpatternconverter.h>
 #include <log4cxx/pattern/formattinginfo.h>
@@ -37,7 +32,7 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  * A flexible layout configurable with pattern string.
  *
  * <p>
- *  The goal of this class is to #format a {@link spi::LoggingEvent LoggingEvent} and
+ *  The goal of this class is to format a {@link spi::LoggingEvent LoggingEvent} and
  *  return the results as a string. The results depend on the <em>conversion pattern</em>.
  * </p>
  *
@@ -62,10 +57,13 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  * <p>
  *  Let the conversion pattern be <strong>"%-5p [%t]: %m%n"</strong> and assume that the log4cxx
  *  environment was set to use a PatternLayout. Then the statements
- *  <pre>
- *      LoggerPtr root = Logger::getRoot();
+ *
+ * ~~~{.cpp}
+ *      auto root = Logger::getRootLogger();
  *      root->debug("Message 1");
- *      root->warn("Message 2");</pre>
+ *      root->warn("Message 2");
+ * ~~~
+ *
  *  would yield the output
  *  <pre>
  *      DEBUG [main]: Message 1
@@ -90,7 +88,7 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  *  <tr>
  *      <td align="center"><strong>c</strong></td>
  *      <td>
- *          Used to output the logger of the logging event. The logger conversion specifier
+ *          Used to output the name of the logger generating the logging event. The <strong>c</strong> conversion specifier
  *          can be optionally followed by <em>precision specifier</em>, that is a decimal
  *          constant in brackets.
  *          <p>
@@ -134,6 +132,12 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  *      </td>
  *  </tr>
  *  <tr>
+ *      <td align="center"><strong>f</strong></td>
+ *      <td>
+ *          Used to output the short file name where the logging request was issued.
+ *      </td>
+ *  </tr>
+ *  <tr>
  *      <td align="center"><strong>F</strong></td>
  *      <td>
  *          Used to output the file name where the logging request was issued.
@@ -155,8 +159,10 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  *  <tr>
  *      <td align="center"><strong>m</strong></td>
  *      <td>
- *          Used to output the application supplied message associated with the logging
- *          event.
+ *          Used to output the application supplied message associated with the logging event.
+ *          To output in a quoted context, add set of braces containing the quote character.
+ *          Any quote character in the message is augmented with a second quote character.
+ *          For example, use %m{'} in an SQL insert statement.
  *      </td>
  *  </tr>
  *  <tr>
@@ -200,8 +206,12 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  *      </td>
  *  </tr>
  *  <tr>
- *      <td align="center"><strong>t</strong></td>
- *      <td>Used to output the name of the thread that generated the logging event.</td>
+ *      <td align="center"><strong>t</strong><p><strong>thread</strong></p></td>
+ *      <td>Used to output the ID of the thread that generated the logging event.</td>
+ *  </tr>
+ *  <tr>
+ *      <td align="center"><strong>T</strong><p><strong>threadname</strong></p></td>
+ *      <td>Used to output the name of the thread that generated the logging event.  May not be available on all platforms.</td>
  *  </tr>
  *  <tr>
  *      <td align="center"><strong>x</strong></td>
@@ -214,14 +224,46 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  *      <td align="center"><strong>X</strong></td>
  *      <td>
  *          Used to output the MDC (mapped diagnostic context) associated with the thread that
- *          generated the logging event. The <strong>X</strong> conversion character <em>must</em> be
- *          followed by the key for the map placed between braces, as in <strong>%X{clientNumber}</strong>
- *          where <code>clientNumber</code> is the key. The value in the MDC corresponding to
+ *          generated the logging event. All key/value pairs are output, each inside <strong>{}</strong> unless
+ *          the <strong>X</strong> is followed by a key placed between braces, as in <strong>%X{clientNumber}</strong>
+ *          where <code>clientNumber</code> is the key. In this case the value in the MDC corresponding to
  *          the key will be output.
  *          <p>See MDC class for more details.</p>
  *      </td>
  *  </tr>
  *  <tr>
+ *      <td align="center"><strong>J</strong></td>
+ *      <td>
+ *          Used to output JSON key/value pairs of all MDC (mapped diagnostic context)
+ *          entries associated with the thread that generated the logging event.
+ *          To output in a quoted context, add set of braces containing the quote character.
+ *          Any quote character in the message is augmented with a second quote character.
+ *          For example, use %J{'} in an SQL insert statement.
+ *          <p>See MDC class for more details.</p>
+ *      </td>
+ *  </tr>
+ *  <tr>
+ *      <td align="center"><strong>y</strong></td>
+ *      <td>
+ *          Used to wrap log with color. The <strong>y</strong> is the end of a color block.<br>
+ *      </td>
+ *  </tr>
+ *  <tr>
+ *      <td align="center"><strong>Y</strong></td>
+ *      <td>
+ *          Used to wrap log with color. The <strong>Y</strong> is the start of a color block.
+ *          Color will be taken from the log level.  The default colors are:
+ *          - `TRACE` - blue
+ *          - `DEBUG` - cyan
+ *          - `INFO` - green
+ *          - `WARN` - yellow
+ *          - `ERROR` - red
+ *          - `FATAL` - magenta
+ *
+ *			These colors are all customizable.
+ *      </td>
+ *  </tr>
+ *  <tr>
  *      <td align="center"><strong>%</strong></td>
  *      <td>The sequence %% outputs a single percent sign.</td>
  *  </tr>
@@ -314,10 +356,10 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  *
  * <p>Below are some examples of conversion patterns.</p>
  *
- * <p><strong>%r [%t] %-5p %c %x - %m\n</strong></p>
+ * <p><strong>%%r [%%t] %-5p %%c %%x - %%m\n</strong></p>
  * <p>This is essentially the TTCC layout.</p>
  *
- * <p><strong>%-6r [%15.15t] %-5p %30.30c %x - %m\n</strong></p>
+ * <p><strong>%-6r [%15.15t] %-5p %30.30c %%x - %%m\n</strong></p>
  *
  * <p>
  *  Similar to the TTCC layout except that the relative time is right padded if less than 6
@@ -330,23 +372,80 @@ LOG4CXX_LIST_DEF(FormattingInfoList, log4cxx::pattern::FormattingInfoPtr);
  *  The above text is largely inspired from Peter A. Darnell and Philip E. Margolis' highly
  *  recommended book "C -- a Software Engineering Approach", ISBN 0-387-97389-3.
  * </p>
+ *
+ * <h2>Colorizing log output</h2>
+ *
+ * In order to customize the colors for the %Y specifier, options have been added to the
+ * PatternLayout in order to let users define their own colors.  Foreground/background colors
+ * can be set, as well as other text effects.  A user can also set the ANSI escape pattern
+ * directly if they want.
+ *
+ * Each level can be set individually.  The configuration options for the levels are as follows:
+ * <ul>
+ * <li>FatalColor</li>
+ * <li>ErrorColor</li>
+ * <li>WarnColor</li>
+ * <li>InfoColor</li>
+ * <li>DebugColor</li>
+ * <li>ErrorColor</li>
+ * </ul>
+ *
+ * Foreground colors may be set by using fg(), background colors may be set by using bg(), and
+ * other options(such as bold) may be set as well.
+ *
+ * Available colors:
+ * <ul>
+ * <li>Black</li>
+ * <li>Red</li>
+ * <li>Green</li>
+ * <li>Yellow</li>
+ * <li>Blue</li>
+ * <li>Magenta</li>
+ * <li>Cyan</li>
+ * <li>White</li>
+ * </ul>
+ *
+ * Available graphics modes:
+ * <ul>
+ * <li>Bold</li>
+ * <li>Dim</li>
+ * <li>Italic</li>
+ * <li>Underline</li>
+ * <li>Blinking</li>
+ * <li>Inverse</li>
+ * <li>Strikethrough</li>
+ * </ul>
+ *
+ * Combining these together, we can configure our colors as we want, shown here in XML:
+ *
+ * Set the background color to red, make text blinking and bold:
+ * <pre>
+ * &lt;param name="FatalColor" value="bg(red)|blinking|bold"/&gt;
+ * </pre>
+ *
+ * Set the foreground color to blue:
+ * <pre>
+ * &lt;param name="FatalColor" value="fg(blue)"/&gt;
+ * </pre>
+ *
+ * Set the foreground color to white and the background color to black:
+ * <pre>
+ * &lt;param name="FatalColor" value="fg(white)|bg(black)"/&gt;
+ * </pre>
+ *
+ * Clear the formatting for the specified level(no formatting will be applied):
+ * <pre>
+ * &lt;param name="FatalColor" value="none"/&gt;
+ * </pre>
+ *
+ * Set a color based off on an ANSI escape sequence(equivalent to setting fg(red)):
+ * <pre>
+ * &lt;param name="FatalColor" value="\x1b[31m"/&gt;
+ * </pre>
  */
 class LOG4CXX_EXPORT PatternLayout : public Layout
 {
-		/**
-		 * Conversion pattern.
-		 */
-		LogString conversionPattern;
-
-		/**
-		 * Pattern converters.
-		 */
-		LoggingEventPatternConverterList patternConverters;
-
-		/**
-		 * Field widths and alignment corresponding to pattern converters.
-		 */
-		FormattingInfoList patternFields;
+		LOG4CXX_DECLARE_PRIVATE_MEMBER_PTR(PatternLayoutPrivate, m_priv)
 
 	public:
 		DECLARE_LOG4CXX_OBJECT(PatternLayout)
@@ -365,6 +464,8 @@ class LOG4CXX_EXPORT PatternLayout : public Layout
 		 */
 		PatternLayout(const LogString& pattern);
 
+		~PatternLayout();
+
 		/**
 		 * Set the <strong>ConversionPattern</strong> option. This is the string which
 		 * controls formatting and consists of a mix of literal content and
@@ -375,24 +476,21 @@ class LOG4CXX_EXPORT PatternLayout : public Layout
 		/**
 		 * Returns the value of the <strong>ConversionPattern</strong> option.
 		 */
-		inline LogString getConversionPattern() const
-		{
-			return conversionPattern;
-		}
+		LogString getConversionPattern() const;
 
 		/**
 		 * Call createPatternParser
 		 */
-		virtual void activateOptions(log4cxx::helpers::Pool& p);
+		void activateOptions(helpers::Pool& p) override;
 
-		virtual void setOption(const LogString& option, const LogString& value);
+		void setOption(const LogString& option, const LogString& value) override;
 
 		/**
 		 * The PatternLayout does not handle the throwable contained within
 		 * {@link spi::LoggingEvent LoggingEvents}. Thus, it returns
 		 * <code>true</code>.
 		 */
-		virtual bool ignoresThrowable() const
+		bool ignoresThrowable() const override
 		{
 			return true;
 		}
@@ -400,19 +498,18 @@ class LOG4CXX_EXPORT PatternLayout : public Layout
 		/**
 		 * Produces a formatted string as specified by the conversion pattern.
 		 */
-		virtual void format(    LogString& output,
+		void format(    LogString& output,
 			const spi::LoggingEventPtr& event,
-			log4cxx::helpers::Pool& pool) const;
+			helpers::Pool& pool) const override;
 
 	protected:
 		virtual log4cxx::pattern::PatternMap getFormatSpecifiers();
+
+	private:
+		pattern::PatternConverterPtr createColorStartPatternConverter(const std::vector<LogString>& options);
 };
 
 LOG4CXX_PTR_DEF(PatternLayout);
 } // namespace log4cxx
 
-#if defined(_MSC_VER)
-	#pragma warning ( pop )
-#endif
-
 #endif //_LOG4CXX_PATTERN_LAYOUT_H
